#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to create a painting on satin

# (c) Dinasset 2021
#
#   History:
#
#   v1.0: 2021-11: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import time
import datetime
from gimpfu import *

# define here some constant value, when required
FixWH=2000

# predefined colour names
NEUTRAL_GRAY=(170,170,170,255)
RED_GRAY=(200,170,170,255)
ORANGE_GRAY=(200,185,170,255)
YELLOW_GRAY=(200,200,170,255)
CHARTREUSE_GRAY=(185,200,170,255)
GREEN_GRAY=(170,200,170,255)
SPRING_GRAY=(170,200,185,255)
CYAN_GRAY=(170,200,200,255)
AZURE_GRAY=(170,185,200,255)
BLUE_GRAY=(170,170,200,255)
VIOLET_GRAY=(185,170,200,255)
MAGENTA_GRAY=(200,170,200,255)
ROSE_GRAY=(200,170,185,255)

############################################################################
# Service routines
#---------------------------------------------------------------------------
  
#---------------------------------------- 
# for messages use: erro_msg or info_msg
#---------------------------------------- 
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

#----------------------------------------------- 
# from Original drawable create the Output image
#----------------------------------------------- 
def createWrkimage(inImage, inLayer):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    OrigWH = inLayer.width+inLayer.height    
    
   # define/calculate the scale factor for the (new) output image  
    scale = (FixWH*1.0)/(OrigWH*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size  
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)                    
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    
    # adjust mean values
    adjustmeanSV (newImage, newBG)
    # cut extremes 
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE,
                                      0.0, 1.0, False, 1.0,
                                      0.15, 0.65, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
   
    newBG.opacity = 100.0
    
    return(newImage)
#---------------------------------------------------------------------------

#---------------------------------------------------------------------------
# adjust values and saturation
#---------------------------------------------------------------------------
def adjustmeanSV (inImage, inLayer):
    # decompose HSV
    (HSVI,HL,SL,VL) = pdb.plug_in_decompose (inImage, inLayer, "HSV", True)
    #pdb.gimp_display_new (HSVI)
    HL=HSVI.layers[0]; SL=HSVI.layers[1]; VL=HSVI.layers[2];
    adjustmeanv (HSVI, SL)
    adjustmeanv (HSVI, VL)
    pdb.plug_in_recompose (HSVI, HL)
    return(inLayer)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # take care of the existing mean in determining the theoretical mean
    meanadj=50
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    thmean = (127.5*meanadj/100+meanl*(100.0-meanadj)/100)   
    #info_msg ("thmean="+str(thmean))
    # range for the loops
    minmean = thmean*1.05  
    maxmean = thmean*1.15  
    #info_msg ("target min mean value="+str(minmean))
    #info_msg ("target max mean value="+str(maxmean))
    
    if meanl<minmean:
        count=0
        while meanl<minmean and count<30:
            #info_msg ("mean source value="+str(meanl))
            pdb.gimp_drawable_levels (inLayer, HISTOGRAM_VALUE,
                                      0.0, 1.0, False, 1.1,
                                      0.0, 1.0, False)
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                          HISTOGRAM_VALUE, 0.0, 1.0)
            count+=1
    elif meanl>maxmean:
        count=0
        while meanl>maxmean and count<30:
            #info_msg ("mean source value="+str(meanl))
            pdb.gimp_drawable_levels (inLayer, HISTOGRAM_VALUE,
                                      0.0, 1.0, False, 0.9,
                                      0.0, 1.0, False)
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                          HISTOGRAM_VALUE, 0.0, 1.0)
            count+=1
    
    return(inLayer)
     
# ------------------------------------------------------
# convert RGB to HSV 
# ------------------------------------------------------

def rgb2hsv(rgb):
    (r,g,b,a) = rgb
    r, g, b = r/255.0, g/255.0, b/255.0
    mx = max(r, g, b)
    mn = min(r, g, b)
    df = mx-mn
    if mx == mn:
        h = 0
    elif mx == r:
        h = (60 * ((g-b)/df) + 360) % 360
    elif mx == g:
        h = (60 * ((b-r)/df) + 120) % 360
    elif mx == b:
        h = (60 * ((r-g)/df) + 240) % 360
    if mx == 0:
        s = 0
    else:
        s = (df/mx)*100
    v = (mx*100)-50
    #info_msg ("hue="+str(h))
    #info_msg ("sat="+str(s))
    #info_msg ("val="+str(v))
    return h, s, v


############################################################################
# Main
#---------------------------------------------------------------------------

def DrawFinePencil (inImage, inLayer,
                  UserOutWH,
                  Preset, predefcol, Colour, Look,
                  doFlatten):
    
    
    # here the logic of the filter
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    OrigWH = inLayer.width+inLayer.height    
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    # create new image to work on
    newImage = createWrkimage (inImage, inLayer)
    width = newImage.width
    height = newImage.height
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    pdb.gimp_image_set_active_layer (newImage, Preprocessed) 
    # Apply engrave_modifie P,0.5,8,0,16,40,0,25,1
    if Preset==0:
        P=0
        Pre="none"
    elif Preset==1:
        P=2
        Pre="B"
    else:
        P=4
        Pre="D"
    EngraveParms=str(P)+',0.5,8,0,16,40,0,25,1'
    pdb.plug_in_gmic_qt(newImage, Preprocessed, 1,2,
    '-v - -engrave_modifie '+EngraveParms+' ')
    Engrave = newImage.layers[0]
    Engrave.mode = LAYER_MODE_LINEAR_BURN   #OVERLAY 
    Engrave.opacity = 90.0  
    Engrave.name = "engrave_modifie "+EngraveParms
    
    #pdb.gimp_display_new (newImage) 
    #return
    
    # Paper  
    Paper = pdb.gimp_layer_new (newImage, width, height, RGBA_IMAGE,
                                "Paper", 50.0, LAYER_MODE_GRAIN_MERGE)
    newImage.add_layer (Paper, 1)
    # colour selection
    if predefcol == 0:
        pdb.gimp_context_set_foreground (Colour) 
        (uR,uG,uB,A)=Colour
        po=" paper specified by user ("+str(int(uR))+","+str(int(uG))+","+str(int(uB))+")"
    if predefcol == 1:
        pdb.gimp_context_set_foreground (NEUTRAL_GRAY); po=" paper NEUTRAL_GRAY"
    if predefcol == 2:
        pdb.gimp_context_set_foreground (RED_GRAY); po=" paper RED_GRAY"
    if predefcol == 3:
        pdb.gimp_context_set_foreground (ORANGE_GRAY); po=" paper ORANGE_GRAY"
    if predefcol == 4:
        pdb.gimp_context_set_foreground (YELLOW_GRAY); po=" paper YELLOW_GRAY"
    if predefcol == 5:
        pdb.gimp_context_set_foreground (CHARTREUSE_GRAY); po=" paper CHARTREUSE_GRAY"
    if predefcol == 6:
        pdb.gimp_context_set_foreground (GREEN_GRAY); po=" paper GREEN_GRAY"
    if predefcol == 7:
        pdb.gimp_context_set_foreground (SPRING_GRAY); po=" paper SPRING_GRAY"
    if predefcol == 8:
        pdb.gimp_context_set_foreground (CYAN_GRAY); po=" paper CYAN_GRAY"
    if predefcol == 9:
        pdb.gimp_context_set_foreground (AZURE_GRAY); po=" paper AZURE_GRAY"
    if predefcol == 10:
        pdb.gimp_context_set_foreground (BLUE_GRAY); po=" paper BLUE_GRAY"
    if predefcol == 11:
        pdb.gimp_context_set_foreground (VIOLET_GRAY); po=" paper VIOLET_GRAY"
    if predefcol == 12:
        pdb.gimp_context_set_foreground (MAGENTA_GRAY); po=" paper MAGENTA_GRAY"
    if predefcol == 13:
        pdb.gimp_context_set_foreground (ROSE_GRAY); po=" paper ROSE_GRAY"
    #pdb.gimp_context_set_foreground (Colour)
    pdb.gimp_edit_fill (Paper, FILL_FOREGROUND)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    if Look==0: # monochromatic
        lk="monochromatic"
        # obtain HSV from Colour RGB to colorize paper
        (R,G,B,A)=pdb.gimp_context_get_foreground()
        hue,sat,val = rgb2hsv ((R,G,B,A))
        #info_msg ("HSV="+str(hue)+","str(sat)+","+str(val))
        # Colorize Source  
        pdb.gimp_drawable_colorize_hsl (Preprocessed, hue, sat, val)
    else: # paper over source
        lk="Pencil on Paper and Source"
        Paper.mode = LAYER_MODE_LINEAR_LIGHT
        Paper.opacity = 80.0
    
    # final step
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "Outcome" 
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
        
    outcome.name = st+oname+" with a draw fine pencil look"
    outcome.name = outcome.name+" (preset="+Pre+"),"
    outcome.name = outcome.name+" ("+po+"),"
    outcome.name = outcome.name+" (look="+lk+")"
     
    # user required output image size
    if UserOutWH==0: UserOutWH=OrigWH
    if UserOutWH != FixWH:
        uRatio = UserOutWH*1.00/FixWH*1.00
        uWidth = int(width*uRatio)
        uHeight = int(height*uRatio)
        #info_msg ("SCALING")   # not during tests, scaling is slow
        pdb.gimp_image_scale (newImage, uWidth, uHeight)   # to user chosen size
     
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="dinasset"
menu="<Image>/Diego/SimpleGraphics/"
copyrightYear="2021"
desc="Draw Fine Pencil"

register(
    "DrawFinePencil",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+
                                   "\n (0=same as input)",
                                   0, (0, 7200, 100)),
        (PF_OPTION, "preset", "Preset", 2, ["*none", "B", "D"]),
        (PF_OPTION, "precol", "Predefined paper colour", 0, ["*user (choose below)",
                                                        "NEUTRAL_GRAY=(170,170,170,255)",
                                                        "RED_GRAY=(200,170,170,255)",
                                                        "ORANGE_GRAY=(200,185,170,255)",
                                                        "YELLOW_GRAY=(200,200,170,255)",
                                                        "CHARTREUSE_GRAY=(185,200,170,255)",
                                                        "GREEN_GRAY=(170,200,170,255)",
                                                        "SPRING_GRAY=(170,200,185,255)",
                                                        "CYAN_GRAY=(170,200,200,255)",
                                                        "AZURE_GRAY=(170,185,200,255)",
                                                        "BLUE_GRAY=(170,170,200,255)",
                                                        "VIOLET_GRAY=(185,170,200,255)",
                                                        "MAGENTA_GRAY=(200,170,200,255)",
                                                        "ROSE_GRAY=(200,170,185,255)"
                                                        ]),
        (PF_COLOR, "pcolor", "Free User paper color"+
                           "\n (use light colours)"+
                           "\n (dft antique white)", (250,235,215,255)),   
        (PF_OPTION, "look", "Outcome look", 0, ["Monochromatic",
                                                "Pencil on Paper and Source"]),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    DrawFinePencil,
    menu=menu
)

main()
